// Copyright 2007-2008 OpenNETCF Consulting, LLC
//
// Permission is hereby granted, free of charge, to any person obtaining 
// a copy of this software and associated documentation files (the "Software"), 
// to deal in the Software without restriction, including without limitation 
// the rights to use, copy, modify, merge, publish, distribute, sublicense, 
// and/or sell copies of the Software, and to permit persons to whom the 
// Software is furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in 
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS 
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
// DEALINGS IN THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.IO;
using OpenNETCF.Media.MP3;
using Microsoft.Win32;
using Microsoft.DirectX.AudioVideoPlayback;
using System.Runtime.InteropServices;
using System.Xml;

namespace AudioTagR
{
    public partial class Form1 : Form
    {
        private static int YELLOW_FOLDER = 0;
        private static int BLUE_FOLDER = 1;
        private static int RED_NOTE = 2;
        private static int BLUE_NOTE = 3;
        private static int BLUE_SQUARE = 4;

        private TreeNode m_lastSelectedNode = null;
        private ID3Tag m_tag;
        private Audio m_audio;
        private TreeNode m_dragSource = null;
        private bool m_inferTrackName = false;
        private bool m_inferAlbum = false;
        private bool m_inferArtist = false;
        private bool m_fileTreeDown = false;
        private TreeNode m_organizerRoot;
        private TreeNode m_organizerClickNode;

        public Form1()
        {
            InitializeComponent();

            fileRoot.KeyDown += new KeyEventHandler(root_KeyDown);
            fileTree.NodeMouseClick += new TreeNodeMouseClickEventHandler(fileTree_NodeMouseClick);

            // get last root folder
            using (RegistryKey key = Registry.CurrentUser.CreateSubKey(@"SOFTWARE\OpenNETCF\AudioTagR"))
            {
                fileRoot.Text = key.GetValue("RootFolder", string.Empty) as string;

                if (Directory.Exists(fileRoot.Text))
                {
                    LoadFileTree();
                }

                m_inferTrackName = (((int)key.GetValue("InferTrack", 1)) != 0);
                m_inferAlbum = (((int)key.GetValue("InferAlbum", 1)) != 0);
                m_inferArtist = (((int)key.GetValue("InferArtist", 1)) != 0);
            }

            m_organizerRoot = new TreeNode("Artists", BLUE_SQUARE, BLUE_SQUARE);
            LoadOrganizer();
            organizerTree.Nodes.Add(m_organizerRoot);

            selectedAlbum.KeyPress += new KeyPressEventHandler(selectedItemKeyPress);
            selectedTrackNum.KeyPress += new KeyPressEventHandler(selectedItemKeyPress);
            selectedTitle.KeyPress += new KeyPressEventHandler(selectedItemKeyPress);
            selectedArtist.KeyPress += new KeyPressEventHandler(selectedItemKeyPress);
            selectedYear.KeyPress += new KeyPressEventHandler(selectedItemKeyPress);

            organizerTree.ContextMenuStrip = null;
            fileTree.ContextMenuStrip = null;
            fileTree.AfterLabelEdit += new NodeLabelEditEventHandler(fileTree_AfterLabelEdit);
        }

        void selectedItemKeyPress(object sender, KeyPressEventArgs e)
        {
            selectedApplyChanges.Enabled = true;
        }

        void fileTree_NodeMouseClick(object sender, TreeNodeMouseClickEventArgs e)
        {

            if (!e.Node.Equals(m_lastSelectedNode))
            {
                selectedAlbum.Text =
                    selectedTrackNum.Text =
                    selectedTitle.Text =
                    selectedArtist.Text =
                    selectedYear.Text = string.Empty;

                selectedApplyChanges.Enabled = false;

                m_tag = null;

                if ((File.Exists(e.Node.Tag as string)) && (string.Compare(Path.GetExtension(e.Node.Tag as string), ".mp3", true) == 0))
                {
                    // it's a file - get the ID3Tag
                    m_tag = ID3Tag.FromFile(e.Node.Tag as string);

                    UpdateFileInfoFromTag(m_tag);
                }
                else if (e.Node.GetNodeCount(false) == 0)
                {
                    AddChildNodes(e.Node);
                }
            }

            m_lastSelectedNode = e.Node;
        }

        void UpdateFileInfoFromTag(ID3Tag tag)
        {
            selectedAlbum.Text = tag.Album;
            selectedTrackNum.Text = tag.TrackNumber;
            selectedTitle.Text = tag.Title;
            selectedArtist.Text = tag.Artist;
            selectedYear.Text = tag.Year;
        }

        void root_KeyDown(object sender, KeyEventArgs e)
        {
            switch (e.KeyCode)
            {
                case Keys.Return:
                    LoadFileTree();
                    break;
            }
        }

        private void LoadFileTree()
        {
            string root = fileRoot.Text.Trim();

            if (!Directory.Exists(root))
            {
                fileRoot.SelectAll();
                fileRoot.Focus();
                return;
            }

            TreeNode rootNode = new TreeNode(root);
            rootNode.Tag = root;

            fileTree.Nodes.Clear();

            fileTree.Nodes.Add(rootNode);

            AddChildNodes(rootNode);
        }

        [DllImport("kernel32.dll", SetLastError=true)]
        private static extern int GetShortPathName(String pathName, StringBuilder shortName, int cbShortName);

        private void AddChildNodes(TreeNode rootNode)
        {
            string root = rootNode.Tag as string;
            string[] dirs;

            try
            {
                dirs = Directory.GetDirectories(root);
            }
            catch (PathTooLongException)
            {
                StringBuilder sb = new StringBuilder(1024);
                root = string.Format("\"{0}\"", root);
                int length = GetShortPathName(root, sb, sb.Capacity);

                dirs = Directory.GetDirectories(sb.ToString());

                if (length == 0) // check for errors
                  MessageBox.Show(Marshal.GetLastWin32Error().ToString());
                else
                  MessageBox.Show(sb.ToString());
            }

            foreach (string d in dirs)
            {
                TreeNode node = new TreeNode(Path.GetFileName(d), YELLOW_FOLDER, YELLOW_FOLDER);
                node.Tag = d;

                rootNode.Nodes.Add(node);
            }

            string[] files = Directory.GetFiles(root, "*.mp3");

            foreach (string f in files)
            {
                TreeNode node = new TreeNode(Path.GetFileName(f), RED_NOTE, RED_NOTE);
                node.Tag = f;

                rootNode.Nodes.Add(node);
            }
            rootNode.Expand();
        }

        private void browseFolder_Click(object sender, EventArgs e)
        {
            using (FolderBrowserDialog fbd = new FolderBrowserDialog())
            {
                fbd.RootFolder = Environment.SpecialFolder.MyComputer;

                fbd.Description = "Select the Root Directory for your music files";

                if (fbd.ShowDialog() == DialogResult.OK)
                {
                    // get last root folder
                    using (RegistryKey key = Registry.CurrentUser.CreateSubKey(@"SOFTWARE\OpenNETCF\AudioTagR"))
                    {
                        key.SetValue("RootFolder", fbd.SelectedPath, RegistryValueKind.String);
                    }

                    fileRoot.Text = fbd.SelectedPath;
                    LoadFileTree();
                }
            }
        }

        private void selectedApplyChanges_Click(object sender, EventArgs e)
        {
            if (m_tag == null) return;

            m_tag.Album = selectedAlbum.Text == string.Empty ? null : selectedAlbum.Text;
            m_tag.TrackNumber = selectedTrackNum.Text == string.Empty ? null : selectedTrackNum.Text;
            m_tag.Title = selectedTitle.Text == string.Empty ? null : selectedTitle.Text;
            m_tag.Artist = selectedArtist.Text == string.Empty ? null : selectedArtist.Text;
            m_tag.Year = selectedYear.Text == string.Empty ? null : selectedYear.Text;

            m_tag.WriteToFile();
            UpdateFileInfoFromTag(m_tag);
            selectedApplyChanges.Enabled = false;
        }

        private void fileTree_DragDrop(object sender, DragEventArgs e)
        {
            m_dragSource = null;

        }

        private void fileTree_DragEnter(object sender, DragEventArgs e)
        {
            if (m_dragSource != null)
            {
                e.Effect = DragDropEffects.Copy;
            }
            else
            {
                e.Effect = DragDropEffects.None;
            }
        }

        private void fileTree_MouseDown(object sender, MouseEventArgs e)
        {
            TreeNode hitNode = fileTree.HitTest(e.Location).Node;
            if (hitNode == null) return;
            
            fileTree.SelectedNode = hitNode;

            switch (GetFileNodeType(hitNode))
            {
                case FileNodeType.Song:
                    fileTree.ContextMenuStrip = fileTreeContextMenu;
                    break;
                case FileNodeType.Root:
                    renameToolStripMenuItem.Visible = false;
                    fileTree.ContextMenuStrip = fileContextMenu;
                    break;
                case FileNodeType.Artist:
                case FileNodeType.Album:
                    renameToolStripMenuItem.Visible = true;
                    fileTree.ContextMenuStrip = fileContextMenu;
                    break;                    
                default:
                    fileTree.ContextMenuStrip = null;
                    break;
            }

            m_fileTreeDown = true;

            m_dragSource = hitNode;
        }

        private FileNodeType GetOrganizerNodeType(TreeNode node)
        {
            if (node.Tag is TrackTag) return FileNodeType.Song;
            if (node.Tag is AlbumTag) return FileNodeType.Album;
            if (node.Tag as string == "Artist") return FileNodeType.Artist;

            return FileNodeType.Root;
        }

        private FileNodeType GetFileNodeType(TreeNode node)        
        {
            if ((!Directory.Exists(node.Tag as string)) && (!(File.Exists(node.Tag as string))))
            {
                node.Remove();
                return FileNodeType.None;
            }

            if (node.Level == 0) return FileNodeType.Root;

            if (Path.HasExtension(node.Tag as string))
            {
                return FileNodeType.Song;
            }

            foreach (string file in Directory.GetFiles(node.Tag as string))
            {
                if (string.Compare(Path.GetExtension(file), ".mp3", true) == 0)
                {
                    return FileNodeType.Album;
                }
            }

            foreach (string folder in Directory.GetDirectories(node.Tag as string))
            {
                foreach (string file in Directory.GetFiles(folder))
                {
                    if (string.Compare(Path.GetExtension(file), ".mp3", true) == 0)
                    {
                        return FileNodeType.Artist;
                    }
                }
            }

            return FileNodeType.None;
        }

        private void fileTree_MouseUp(object sender, MouseEventArgs e)
        {
            m_fileTreeDown = false;
            m_dragSource = null;
        }

        private void fileTree_MouseLeave(object sender, EventArgs e)
        {
            m_fileTreeDown = false;
        }

        private void fileTree_MouseMove(object sender, MouseEventArgs e)
        {
            if ((m_fileTreeDown) && (m_dragSource != null))
            {
                this.DoDragDrop(m_dragSource, DragDropEffects.Copy);
            }

        }

        private void organizerTree_DragEnter(object sender, DragEventArgs e)
        {
            if (m_dragSource != null)
            {
                e.Effect = DragDropEffects.Copy;
            }
            else
            {
                e.Effect = DragDropEffects.None;
            }
        }

        private bool DoWork(string path, TreeNode artistNode, TreeNode albumNode, TreeNode songNode, int dialogTop, int dialogLeft)
        {
            ID3Tag tag = ID3Tag.FromFile(path);

            string artist = tag.Artist;
            string album = tag.Album;
            string year = tag.Year;
            string title = tag.Title;
            string trackNumber = tag.TrackNumber;

            bool needArtist = (artist == tag.EmptyTagString);
            bool needAlbum = (album == tag.EmptyTagString);
            bool needYear = (year == tag.EmptyTagString);
            bool needTitle = (title == tag.EmptyTagString);
            bool needTrackNumber = (title == tag.EmptyTagString);

            if (!needArtist && (artistNode != null) && (artistNode.Text != artist))
            {
                string question = string.Format("Do you want to change the Artist from '{0}' to '{1}'?", artist, artistNode.Text);
                if (MessageBox.Show(question, "Change Track Info", MessageBoxButtons.YesNo, MessageBoxIcon.Question, MessageBoxDefaultButton.Button2) == DialogResult.No)
                {
                    return false;
                }

                artist = artistNode.Text;
            }

            if (!needAlbum && (albumNode != null) && (albumNode.Text != album))
            {
                string question = string.Format("Do you want to change the Album from '{0}' to '{1}'?", album, albumNode.Text);
                if (MessageBox.Show(question, "Change Track Info", MessageBoxButtons.YesNo, MessageBoxIcon.Question, MessageBoxDefaultButton.Button2) == DialogResult.No)
                {
                    return false;
                }

                album = albumNode.Text;
            }

            if (needArtist && m_inferArtist)
            {
                artist = InferArtistFromFile(path);
                needArtist = false;
            }

            if (needAlbum && m_inferAlbum)
            {
                album = InferAlbumFromFile(path);
                needAlbum = false;
            }

            if (needTitle && m_inferTrackName)
            {
                title = InferTrackNameFromFile(path);
                needTitle = false;
            }

            if (needTrackNumber && m_inferTrackName)
            {
                trackNumber = InferTrackNumberFromFile(path);
                if (trackNumber != "0") needTrackNumber = false;
            }

            if ((needYear) && (!needArtist) && (!needAlbum))
            {
                artistNode = m_organizerRoot.Nodes[artist];
                if (artistNode != null)
                {
                    albumNode = artistNode.Nodes[album];
                    if (albumNode != null)
                    {
                        year = ((AlbumTag)albumNode.Tag).Year;
                    }
                }
            }
            needYear = (year == tag.EmptyTagString);

            // get all the needed info
            // do we need any info?
            bool needSomething = (needArtist || needAlbum || needYear || needTitle || needTrackNumber);

            if (needSomething)
            {
                NewItem ni = new NewItem(needArtist, needAlbum, needYear, needTitle, needTrackNumber);
                ni.Top = dialogTop;
                ni.Left = dialogLeft;
                if (ni.ShowDialog() == DialogResult.Cancel) return false;

                // collect data from input form
                if (needArtist) artist = ni.artist.Text;
                if (needAlbum) album = ni.album.Text;
                if (needYear) year = ni.year.Text;
                if (needTitle) title = ni.title.Text;
                if (needTrackNumber) trackNumber = ni.trackNumber.Text;
            }

            artistNode = m_organizerRoot.Nodes[artist];
            if (artistNode == null)
            {
                artistNode = new TreeNode(artist, BLUE_FOLDER, BLUE_FOLDER);
                artistNode.Tag = "Artist";
                artistNode.Name = artist;
                m_organizerRoot.Nodes.Add(artistNode);
            }

            albumNode = artistNode.Nodes[album];
            if (albumNode == null)
            {
                albumNode = new TreeNode(album, BLUE_FOLDER, BLUE_FOLDER);
                albumNode.Tag = new AlbumTag(album, year);
                albumNode.Name = album;
                artistNode.Nodes.Add(albumNode);
            }
            else
            {
                // we inherit the album year - period
                year = ((AlbumTag)albumNode.Tag).Year;
            }

            songNode = albumNode.Nodes[title];
            if (songNode == null)
            {
                songNode = new TreeNode(title, BLUE_NOTE, BLUE_NOTE);
                songNode.Tag = new TrackTag(path, title, trackNumber);
                songNode.Name = title;
                albumNode.Nodes.Add(songNode);
            }

            songNode.EnsureVisible();

            // stamp file with tag info
            tag.Artist = artist;
            tag.Album = album;
            tag.Year = year;
            tag.Title = title;
            tag.TrackNumber = trackNumber;

            tag.WriteToFile();

            // update file info
            if (File.Exists(fileTree.SelectedNode.Tag as string))
            {
                UpdateFileInfoFromTag(tag);
            }

            return true;
        }

        private void organizerTree_DragDrop(object sender, DragEventArgs e)
        {
            Point p = organizerTree.PointToClient(new Point(e.X, e.Y));
            TreeNode destinationNode = organizerTree.HitTest(p).Node;
            TreeNode artistNode = null;
            TreeNode albumNode = null;
            TreeNode songNode = null;
            bool changeMade = false;

            int dialogTop = this.Top + (SystemInformation.CaptionHeight
                                    + SystemInformation.MenuHeight
                                    + groupBox1.Top
                                    + organizerTree.Top
                                    + destinationNode.Bounds.Top
                                    + destinationNode.Bounds.Height);
            int dialogLeft = this.Left + (groupBox1.Left
                                    + organizerTree.Left
                                    + destinationNode.Bounds.Left);

            if (destinationNode != null)
            {
                string path = (string)m_dragSource.Tag;

                if (Path.HasExtension(path))
                {
                    try
                    {
                        using (FileStream fs = File.OpenWrite(path)) { }
                    }
                    catch
                    {
                        // file is in use
                        MessageBox.Show("File is in use.", "Error");
                        return;
                    }

                    // file drop
                    switch (destinationNode.Level)
                    {
                        case 0:
                            // root node
                            changeMade = DoWork(path, artistNode, albumNode, songNode, dialogTop, dialogLeft);
                            break;
                        case 1:
                            // artist tag
                            artistNode = destinationNode;
                            changeMade = DoWork(path, artistNode, albumNode, songNode, dialogTop, dialogLeft);
                            break;
                        case 2:
                            // album
                            artistNode = destinationNode.Parent;
                            albumNode = destinationNode;
                            changeMade = DoWork(path, artistNode, albumNode, songNode, dialogTop, dialogLeft);
                            break;
                        default:
                            // anything else is invalid
                            break;
                    }
                }
                else
                {
                    // source was a folder
                    if (IsArtistFolder(path))
                    {
                        string[] dirs = Directory.GetDirectories(path);

                        foreach (string dir in dirs)
                        {
                            if (IsAlbumFolder(dir))
                            {
                                changeMade = HandleAlbumFolder(dir, destinationNode, artistNode, albumNode, songNode, dialogTop, dialogLeft);
                            }
                        }
                    }
                    else if (IsAlbumFolder(path))
                    {
                        changeMade = HandleAlbumFolder(path, destinationNode, artistNode, albumNode, songNode, dialogTop, dialogLeft);
                    }
                }
            }

            if (changeMade) SaveOrganizer();
            m_dragSource = null;
        }

        private bool HandleAlbumFolder(string path, TreeNode destinationNode, TreeNode artistNode, TreeNode albumNode, TreeNode songNode, int dialogTop, int dialogLeft)
        {
            bool success = true;

            switch (destinationNode.Level)
            {
                case 1:
                    // artist tag
                    artistNode = destinationNode;
                    break;
                case 2:
                    // album
                    artistNode = destinationNode.Parent;
                    albumNode = destinationNode;
                    break;
            }
            string[] files = Directory.GetFiles(path);

            foreach (string file in files)
            {
                if (string.Compare(Path.GetExtension(file), ".mp3", true) == 0)
                {
                    success = DoWork(file, artistNode, albumNode, songNode, dialogTop, dialogLeft) && success;
                }
            }

            return success;
        }

        private bool IsArtistFolder(string path)
        {
            string[] dirs = Directory.GetDirectories(path);

            foreach (string dir in dirs)
            {
                if (IsAlbumFolder(dir))
                {
                    return true;
                }
            }
            return false;
        }

        private bool IsAlbumFolder(string path)
        {
            string[] files = Directory.GetFiles(path);

            foreach(string file in files)
            {
                if (string.Compare(Path.GetExtension(file), ".mp3", true) == 0)
                {
                    return true;
                }
            }
            return false;
        }

        private string InferArtistFromFile(string path)
        {
            return Path.GetFileName(Path.GetDirectoryName(Path.GetDirectoryName(path)));
        }

        private string InferAlbumFromFile(string path)
        {
            return Path.GetFileName(Path.GetDirectoryName(path));
        }

        private string InferTrackNumberFromFile(string path)
        {
            string filename = Path.GetFileName(path);

            // find the first space, then convert the beginning to an int - failure == 00
            int index = filename.IndexOf(' ');
            if (index <= 0)
            {
                return "0";
            }
            else
            {
                try
                {
                    return int.Parse(filename.Substring(0, index)).ToString();
                }
                catch
                {
                    return "0";
                }
            }
        }

        private string InferTrackNameFromFile(string path)
        {
            string filename = Path.GetFileName(path);

            // find the first space, then convert the beginning to an int - failure == 00
            int index = filename.IndexOf(' ');
            string songTitle = filename.Substring(index + 1);
            return Path.GetFileNameWithoutExtension(songTitle);
        }

        private void organizerTree_MouseDown(object sender, MouseEventArgs e)
        {
            m_organizerClickNode = organizerTree.HitTest(e.Location).Node;

            if (m_organizerClickNode == null) return;

            m_organizerClickNode.TreeView.SelectedNode = m_organizerClickNode;

            switch (GetOrganizerNodeType(m_organizerClickNode))
            {
                case FileNodeType.Root:
                    addItemToolStripMenuItem.Text = "Add New Artist";
                    addItemToolStripMenuItem.Visible = true;
                    removeItemToolStripMenuItem.Visible = false;
                    organizerTree.ContextMenuStrip = organizerTreeContextMenu;
                    break;
                case FileNodeType.Artist:
                    addItemToolStripMenuItem.Text = "Add New Album";
                    removeItemToolStripMenuItem.Text = "Remove Artist";
                    addItemToolStripMenuItem.Visible = true;
                    removeItemToolStripMenuItem.Visible = true;
                    organizerTree.ContextMenuStrip = organizerTreeContextMenu;

                    organizerTabs.SelectedTab = artistTab;
                    organizerArtist.Text = m_organizerClickNode.Text;
                    organizerAlbum.Text = string.Empty;
                    organizerYear.Text = string.Empty;
                    organizerTrackNumber.Text = string.Empty;
                    organizerTrackName.Text = string.Empty;
                    break;
                case FileNodeType.Album:
                    removeItemToolStripMenuItem.Text = "Remove Album";
                    addItemToolStripMenuItem.Visible = false;
                    removeItemToolStripMenuItem.Visible = true;

                    organizerTabs.SelectedTab = albumTab;
                    organizerArtist.Text = m_organizerClickNode.Parent.Text;
                    organizerAlbum.Text = m_organizerClickNode.Text;
                    organizerYear.Text = ((AlbumTag)m_organizerClickNode.Tag).Year;
                    organizerTrackNumber.Text = string.Empty;
                    organizerTrackName.Text = string.Empty;
                    break;
                case FileNodeType.Song:
                    removeItemToolStripMenuItem.Text = "Remove Song";
                    addItemToolStripMenuItem.Visible = false;
                    removeItemToolStripMenuItem.Visible = true;

                    organizerTabs.SelectedTab = trackTab;
                    //TODO
                    organizerArtist.Text = m_organizerClickNode.Parent.Parent.Text;
                    organizerAlbum.Text = m_organizerClickNode.Parent.Text;
                    organizerYear.Text = ((AlbumTag)m_organizerClickNode.Parent.Tag).Year; ;
                    organizerTrackNumber.Text = ((TrackTag)m_organizerClickNode.Tag).TrackNumber; ;
                    organizerTrackName.Text = ((TrackTag)m_organizerClickNode.Tag).SongTitle;
                    break;
                default:
                    organizerTree.ContextMenuStrip = null;
                    organizerArtist.Text = string.Empty;
                    organizerAlbum.Text = string.Empty;
                    organizerYear.Text = string.Empty;
                    organizerTrackNumber.Text = string.Empty;
                    organizerTrackName.Text = string.Empty;
                    break;
            }
        }

        private void removeItemToolStripMenuItem_Click(object sender, EventArgs e)
        {
            ToolStripDropDownItem item = (ToolStripDropDownItem)sender;
            string question = string.Empty;

            if (((ToolStripDropDownItem)sender).Text.Contains("Song"))
            {
                question = string.Format("Remove track '{0}'?", organizerTree.SelectedNode.Text);
            }
            else if (((ToolStripDropDownItem)sender).Text.Contains("Album"))
            {
                question = string.Format("Remove album '{0}' and all tracks it contains?", organizerTree.SelectedNode.Text);
            }
            else if (((ToolStripDropDownItem)sender).Text.Contains("Artist"))
            {
                question = string.Format("Remove artist '{0}' and all of its albums and tracks?", organizerTree.SelectedNode.Text);
            }

            if (question == string.Empty) return;

            if (MessageBox.Show(question, "Remove Organizer Item", MessageBoxButtons.YesNo, MessageBoxIcon.Question, MessageBoxDefaultButton.Button2) == DialogResult.Yes)
            {
                organizerTree.SelectedNode.Remove();
                SaveOrganizer();
            }
        }

        private void addItemToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (addItemToolStripMenuItem.Text.Contains("Artist"))
            {
                using (NewItem ni = new NewItem(true, false, false, false, false))
                {
                    ni.Left = organizerTreeContextMenu.Left;
                    ni.Top = organizerTreeContextMenu.Top;
                    if (ni.ShowDialog() == DialogResult.OK)
                    {
                        TreeNode node = new TreeNode(ni.artist.Text, BLUE_FOLDER, BLUE_FOLDER);
                        node.Name = node.Text;
                        node.Tag = "Artist";

                        if (organizerTree.Nodes[node.Text] == null)
                        {
                            m_organizerClickNode.Nodes.Add(node);
                            node.EnsureVisible();
                        }
                    }
                }
            }
            else if (addItemToolStripMenuItem.Text.Contains("Album"))
            {
                using (NewItem ni = new NewItem(false, true, true, false, false))
                {
                    ni.Left = organizerTreeContextMenu.Left;
                    ni.Top = organizerTreeContextMenu.Top;
                    if (ni.ShowDialog() == DialogResult.OK)
                    {
                        TreeNode node = new TreeNode(ni.album.Text, BLUE_FOLDER, BLUE_FOLDER);
                        node.Name = node.Text;
                        node.Tag = new AlbumTag(ni.album.Text, ni.year.Text);

                        if (organizerTree.Nodes[node.Text] == null)
                        {
                            m_organizerClickNode.Nodes.Add(node);
                            node.EnsureVisible();
                        }
                    }
                }
            }
        }

        private void clearAllSongInfoToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (MessageBox.Show("Are you certain you want to clear all information for this track?",
                "Clear Info",
                MessageBoxButtons.YesNo,
                MessageBoxIcon.Question,
                MessageBoxDefaultButton.Button2) == DialogResult.No)
            {
                return;
            }

            ID3Tag tag = ID3Tag.FromFile(fileTree.SelectedNode.Tag as string);
            tag.Artist = null;
            tag.Album = null;
            tag.Publisher = null;
            tag.Title = null;
            tag.TrackNumber = null;
            tag.Year = null;

            tag.WriteToFile();
            UpdateFileInfoFromTag(tag);
        }

        private void playSongToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (m_audio == null)
            {
                m_audio = new Audio(fileTree.SelectedNode.Tag as string);
                m_audio.Ending += new EventHandler(m_audio_Ending);
                m_audio.Play();
                playSongToolStripMenuItem.Text = "Stop Playing";
            }
            else
            {
                if (m_audio.Playing)
                {
                    m_audio.Stop();
                    playSongToolStripMenuItem.Text = "Play Song";
                }
                else
                {
                    m_audio.Open(fileTree.SelectedNode.Tag as string, true);
                    playSongToolStripMenuItem.Text = "Stop Playing";
                }
            }
        }

        void m_audio_Ending(object sender, EventArgs e)
        {
            playSongToolStripMenuItem.Text = "Play Song";
        }

        private void optionsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            using (Options options = new Options())
            {
                if (options.ShowDialog() == DialogResult.OK)
                {
                    m_inferTrackName = options.inferTrack.Checked;
                    m_inferAlbum = options.inferAlbum.Checked;
                    m_inferArtist = options.inferArtist.Checked;

                    using (RegistryKey key = Registry.CurrentUser.CreateSubKey(@"SOFTWARE\OpenNETCF\AudioTagR"))
                    {
                        key.SetValue("InferTrack", m_inferTrackName ? 1 : 0, RegistryValueKind.DWord);
                        key.SetValue("InferAlbum", m_inferAlbum ? 1 : 0, RegistryValueKind.DWord);
                        key.SetValue("InferArtist", m_inferArtist ? 1 : 0, RegistryValueKind.DWord);
                        key.SetValue("LookupURL", options.lookupUrl.Text, RegistryValueKind.String);
                    }
                }
            }
        }

        private void aboutToolStripMenuItem_Click(object sender, EventArgs e)
        {
            MessageBox.Show("Tony Tacke's AudioTagR\r\nChristmas 2007\r\nHo Ho Ho!!", "About AudioTagR"); 
        }

        private void LoadOrganizer()
        {
            string path = Path.Combine(Application.StartupPath, "AudioTagR.xml");
            XmlDocument doc = new XmlDocument();

            if (!File.Exists(path)) return;

            doc.Load(path);
            XmlNode xmlroot = doc.SelectSingleNode("Artists");
            foreach (XmlNode artistXml in xmlroot.SelectNodes("Artist"))
            {
                TreeNode artistNode = new TreeNode(artistXml.Attributes["Name"].Value, BLUE_FOLDER, BLUE_FOLDER);
                artistNode.Tag = "Artist";
                artistNode.Name = artistNode.Text;

                XmlNode albumsxml = artistXml.SelectSingleNode("Albums");
                foreach (XmlNode albumXml in albumsxml.SelectNodes("Album"))
                {
                    AlbumTag atag = new AlbumTag(albumXml.Attributes["Name"].Value, albumXml.Attributes["Year"].Value);
                    TreeNode albumNode = new TreeNode(albumXml.Attributes["Name"].Value, BLUE_FOLDER, BLUE_FOLDER);
                    albumNode.Name = albumNode.Text;
                    albumNode.Tag = atag;

                    XmlNode tracksxml = albumXml.SelectSingleNode("Tracks");
                    foreach (XmlNode trackXml in tracksxml.SelectNodes("Track"))
                    {
                        TrackTag ttag = new TrackTag(string.Empty, trackXml.Attributes["Title"].Value, trackXml.Attributes["Number"].Value);
                        TreeNode trackNode = new TreeNode(trackXml.Attributes["Title"].Value, BLUE_NOTE, BLUE_NOTE);
                        trackNode.Tag = ttag;
                        albumNode.Nodes.Add(trackNode);
                    }

                    artistNode.Nodes.Add(albumNode);
                }
                m_organizerRoot.Nodes.Add(artistNode);
                m_organizerRoot.Expand();
            }
        }

        private void SaveOrganizer()
        {
            string path = Path.Combine(Application.StartupPath, "AudioTagR.xml");
      
            using (XmlTextWriter writer = new XmlTextWriter(path, Encoding.ASCII))
            {
                writer.Formatting = Formatting.Indented;

                writer.WriteStartDocument();
                writer.WriteStartElement("Artists");

                foreach(TreeNode artistNode in m_organizerRoot.Nodes)
                {
                    writer.WriteStartElement("Artist");
                    writer.WriteAttributeString("Name", artistNode.Text);

                    writer.WriteStartElement("Albums");
                    foreach (TreeNode albumNode in artistNode.Nodes)
                    {
                        AlbumTag atag = (AlbumTag)albumNode.Tag;
                        writer.WriteStartElement("Album");
                        writer.WriteAttributeString("Name", atag.Name);
                        writer.WriteAttributeString("Year", atag.Year);

                        writer.WriteStartElement("Tracks");
                        foreach (TreeNode trackNode in albumNode.Nodes)
                        {
                            TrackTag ttag = (TrackTag)trackNode.Tag;
                            writer.WriteStartElement("Track");
                            writer.WriteAttributeString("Title", ttag.SongTitle);
                            writer.WriteAttributeString("Number", ttag.TrackNumber);
                            writer.WriteEndElement(); // track
                        }

                        writer.WriteEndElement(); // tracks
                        writer.WriteEndElement(); // album
                    }
                    writer.WriteEndElement(); // Albums
                    writer.WriteEndElement(); // Artist
                }
                writer.WriteEndElement(); // Arists
            }
        }

        private string m_lastLabel = null;

        private void renameToolStripMenuItem_Click(object sender, EventArgs e)
        {
            m_lastLabel = fileTree.SelectedNode.Text;
            fileTree.LabelEdit = true;
            fileTree.SelectedNode.BeginEdit();
        }

        void fileTree_AfterLabelEdit(object sender, NodeLabelEditEventArgs e)
        {
            try
            {
                if (e.Label == m_lastLabel)
                {
                    e.CancelEdit = true;
                    return;
                }

                string dir = e.Node.Tag as string;
                if(dir == null)
                {
                    e.CancelEdit = true;
                    return;
                }

                // try to rename the folder.
                if (!Directory.Exists(dir))
                {
                    e.CancelEdit = true;
                    return;
                }

                try
                {
                    string newDir = Path.GetDirectoryName(dir);
                    newDir = Path.Combine(newDir, e.Label);
                    Directory.Move(dir, newDir);
                    e.Node.Tag = newDir;
                }
                catch(Exception ex)
                {
                    MessageBox.Show("Unable to rename folder:\r\n" + ex.Message, "Error");
                    e.CancelEdit = true;
                    return;
                }
                e.Node.Nodes.Clear();
                AddChildNodes(e.Node);
            }
            finally
            {
                m_lastLabel = null;
                fileTree.LabelEdit = false;
            }
            fileTree.LabelEdit = false;
        }

        private void refreshToolStripMenuItem_Click(object sender, EventArgs e)
        {
            TreeNode node = fileTree.SelectedNode;

            if (node == null) return;

            node.Nodes.Clear();
            AddChildNodes(node);            
        }

    }
}