// Copyright 2007-2012 OpenNETCF Consulting, LLC
//
// Permission is hereby granted, free of charge, to any person obtaining 
// a copy of this software and associated documentation files (the "Software"), 
// to deal in the Software without restriction, including without limitation 
// the rights to use, copy, modify, merge, publish, distribute, sublicense, 
// and/or sell copies of the Software, and to permit persons to whom the 
// Software is furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in 
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS 
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
// DEALINGS IN THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Text;
using OpenNETCF.Media.MP3.Frames;

namespace OpenNETCF.Media.MP3
{
    internal class FrameCollection : IEnumerable<ID3Frame>
    {
        private List<ID3Frame> m_frames = new List<ID3Frame>();
        private bool m_dirty = false;
        private string m_emptyTagString = "<Unknown>";

        internal void Add(ID3Frame frame)
        {
            m_frames.Add(frame);
        }

        public ID3Frame this[int index]
        {
            get { return m_frames[index]; }
        }

        public string EmptyTagString
        {
            get { return m_emptyTagString; }
            set { m_emptyTagString = value; }
        }

        public int Count
        {
            get { return m_frames.Count; }
        }

        public IEnumerator<ID3Frame> GetEnumerator()
        {
            return m_frames.GetEnumerator();
        }

        System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
        {
            return m_frames.GetEnumerator();
        }

        public T GetFrame<T>() where T : ID3Frame
        {
            foreach (ID3Frame frame in this)
            {
                T t = frame as T;
                if (t != null) return t;
            }
            return null;
        }

        public void RemoveFrame<T>() where T : ID3Frame
        {
            for (int i = m_frames.Count - 1 ; i >= 0; i--)
            {
                T t = m_frames[i] as T;
                if (t != null)
                {
                    m_frames.RemoveAt(i);
                    m_dirty = true;
                }
            }
        }

        public string GetFrameText<T>() where T : TextFrame
        {
            T frame = GetFrame<T>();
            return (frame == null) ? m_emptyTagString : frame.Text;
        }

        public void SetFrameText<T>(string text) where T : TextFrame, new()
        {
            T frame = GetFrame<T>();
            if (frame == null)
            {
                // we have to add this frame
                frame = new T();
                frame.Text = text;
                this.Add(frame);
            }
            else
            {
                if (frame.Text == text)
                {
                    // nothing to do
                    return;
                }

                // the frame exists, we just need to set its value
                frame.Text = text;
            }
            m_dirty = true;
        }

        public bool HasChanged
        {
            get { return m_dirty; }
        }

        public byte[] GetBytes()
        {
            List<byte> data = new List<byte>();

            foreach (ID3Frame frame in m_frames)
            {
                data.AddRange(frame.GetBytes());
            }

            return data.ToArray();
        }
    }
}
