// Copyright 2007-2012 OpenNETCF Consulting, LLC
//
// Permission is hereby granted, free of charge, to any person obtaining 
// a copy of this software and associated documentation files (the "Software"), 
// to deal in the Software without restriction, including without limitation 
// the rights to use, copy, modify, merge, publish, distribute, sublicense, 
// and/or sell copies of the Software, and to permit persons to whom the 
// Software is furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in 
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS 
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
// DEALINGS IN THE SOFTWARE.

using System;
using System.IO;
using System.Text;
using System.Collections.Generic;

namespace OpenNETCF.Media.MP3.Frames
{
    internal abstract class ID3Frame
    {
        private const int ID_OFFSET = 0;
        private const int ID_LENGTH = 4;
        private const int SIZE_OFFSET = ID_OFFSET + ID_LENGTH;
        private const int SIZE_LENGTH = 4;
        private const int STATUS_FLAGS_OFFSET = SIZE_OFFSET + SIZE_LENGTH;
        private const int STATUS_FLAGS_LENGTH = 1;
        private const int FORMAT_FLAGS_OFFSET = STATUS_FLAGS_OFFSET + STATUS_FLAGS_LENGTH;
        private const int FORMAT_FLAGS_LENGTH = 1;

        protected const int HEADER_LENGTH = FORMAT_FLAGS_OFFSET + FORMAT_FLAGS_LENGTH;

        protected byte[] m_data;

        protected ID3Frame(byte[] data)
        {
            m_data = data;
        }

        public byte[] GetBytes()
        {
            return m_data;
        }

        internal static byte[] GetFrameBytesFromStream(FileStream stream, int maxPos)
        {
            // pull the header so we can find the length
            byte[] header = new byte[HEADER_LENGTH];
            stream.Read(header, 0, header.Length);

            // get the data length
            int dataSize = (new SynchSafeInteger(header, SIZE_OFFSET)).ToInt32();

            // would this put us past the header (i.e. it's not really a frame)?
            if ((stream.Position + dataSize) > maxPos)
            {
                // back up
                stream.Seek(-HEADER_LENGTH, SeekOrigin.Current);

                // return nothing
                return null;
            }

            byte[] data = new byte[dataSize];

            // read the frame data
            stream.Read(data, 0, dataSize);

            // copy everything to our internal buffer
            byte[] frameData = new byte[header.Length + data.Length];
            Buffer.BlockCopy(header, 0, frameData, 0, header.Length);
            Buffer.BlockCopy(data, 0, frameData, header.Length, data.Length);

            return frameData;
        }

        internal static string IDFromBytes(byte[] data)
        {
            return Encoding.ASCII.GetString(data, ID_OFFSET, ID_LENGTH);
        }

        public FrameFormatFlags FormatFlags
        {
            get { return (FrameFormatFlags)m_data[FORMAT_FLAGS_OFFSET]; }
        }

        public FrameStatusFlags StatusFlags
        {
            get { return (FrameStatusFlags)m_data[STATUS_FLAGS_OFFSET]; }
        }

        public string ID
        {
            get { return Encoding.ASCII.GetString(m_data, ID_OFFSET, ID_LENGTH); }
            set
            {
                if (value.Length != 4) throw new Exception("Invalid ID");

                byte[] textBytes = Encoding.ASCII.GetBytes(value);
                Buffer.BlockCopy(textBytes, 0, m_data, ID_OFFSET, ID.Length);
            }
        }

        protected void UpdateLength(int length)
        {
            SynchSafeInteger ssi = new SynchSafeInteger(length);
            Buffer.BlockCopy(ssi.GetBytes(), 0, m_data, SIZE_OFFSET, SIZE_LENGTH);
        }
    }
}
