// Copyright 2007-2012 OpenNETCF Consulting, LLC
//
// Permission is hereby granted, free of charge, to any person obtaining 
// a copy of this software and associated documentation files (the "Software"), 
// to deal in the Software without restriction, including without limitation 
// the rights to use, copy, modify, merge, publish, distribute, sublicense, 
// and/or sell copies of the Software, and to permit persons to whom the 
// Software is furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in 
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS 
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
// DEALINGS IN THE SOFTWARE.

using System;
using System.IO;

namespace OpenNETCF.Media.MP3
{
    internal sealed class ID3Header
    {
        private const int ID_OFFSET = 0;
        private const int ID_LENGTH = 3;
        private const int VERSION_OFFSET = ID_OFFSET + ID_LENGTH;
        private const int VERSION_LENGTH = 2;
        private const int FLAGS_OFFSET = VERSION_OFFSET + VERSION_LENGTH;
        private const int FLAGS_LENGTH = 1;
        private const int SIZE_OFFSET = FLAGS_OFFSET + FLAGS_LENGTH;
        private const int SIZE_LENGTH = 4;
        public const int HEADER_LENGTH = SIZE_OFFSET + SIZE_LENGTH;

        private byte[] m_data = new byte[HEADER_LENGTH];

        public byte[] GetBytes()
        {
            return m_data;
        }
        
        private ID3Header(FileStream stream)
        {
            stream.Read(m_data, 0, HeaderLength);
        }

        private ID3Header(byte[] data)
        {
            Buffer.BlockCopy(data, 0, m_data, 0, HeaderLength);
        }

        public static ID3Header FromStream(FileStream stream)
        {
            return new ID3Header(stream);
        }

        public static ID3Header FromBytes(byte[] data)
        {
            return new ID3Header(data);
        }

        public static ID3Header GetEmptyHeader()
        {
            byte[] data = new byte[] { 0x49, 0x44, 0x33, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
            return new ID3Header(data);
        }

        public static int HeaderLength
        {
            get { return HEADER_LENGTH; }
        }

        public int Version
        {
            get { return (int)m_data[VERSION_OFFSET]; }
        }

        public int Length
        {
            get { return HEADER_LENGTH; }
        }

        public ID3HeaderFlags Flags
        {
            get { return (ID3HeaderFlags)m_data[FLAGS_OFFSET]; }
        }

        public int TagLength
        {
            get { return (new SynchSafeInteger(m_data, SIZE_OFFSET)).ToInt32(); }
            set 
            {
                SynchSafeInteger ssi = new SynchSafeInteger(value);
                Buffer.BlockCopy(ssi.GetBytes(), 0, m_data, SIZE_OFFSET, 4);
            }
        }
    }
}
